
#ifndef CFSRW_H
#define CFSRW_H

#include "arpc.h"
#include "sfsclient.h"
#include "sfsmisc.h"
#include "dhash_prot.h"
#include "dhash.h"
#include "vervec.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "cfsa_prot.h"
#include "scan.h"
#include "snapshot.h"

struct session {
  nfscall *sbp;
  ptr<materialized_view> mv;
  ptr<log> l;
  vervec stop;
  ptr<snapshot> image;
};
typedef ptr<session> sessionp;

class cfsrw
{
private:
  struct snaprec {
    bool set;
    bool building;
    unsigned ts;
    chordID location;
    vervec version;
    ptr<snapshot> image;
  };

  ref<view> _view;
  chordID _vid;
  bigint& _verf;

  vervec _version;
  qhash<sfs_aid, ptr<log> > _logs;
  ptr<log> mylog(nfscall*);
  ptr<log> mylog (sfs_aid);

  qhash<bigint, bool> _locks;
  qhash<chordID, ptr<snaprec> > _snapshots;

  callback<void, nfscall *, void *, nfs_fh3>::ref getreply;
  
  void reply (nfscall *sbp, void *res)
  {
    nfs_fh3 fh;
    cid2nfs (0, &fh);
    getreply (sbp, res, fh);
  }
  
  void reply (nfscall *sbp, void *res, chordID a)
  {
    nfs_fh3 fh;
    cid2nfs (a, &fh);
    getreply (sbp, res, fh);
  }

  void echo (recordp, cb_recordp, bool);
  void noop_bool (bool);
  void noop_recordp (recordp);
  ptr<inode_log_record> make_inode (sessionp, int, cb_recordp);

  void getattr_check_snapshot (sessionp s, chordID i, cb_fa cb,
                               recordp e, ptr<cattr> ca, nfsstat3 status);
  void getattr_inode_cb (sessionp s, chordID i, cb_fa cb,
                         recordp e, ptr<cattr> ca, nfsstat3 status);
 
  void find_inode_check_snapshot (sessionp, chordID, cb_inode, recordp,
                                  ptr<cattr>, nfsstat3);
  void lookup_check_snapshot (sessionp, chordID, str, cb_lookup,
                              recordp, nfsstat3);
  void read_check_snapshot (sessionp, cb_read cb, ptr<read_state>,
                            nfsstat3);
  bool readdir_check_snapshot (sessionp, chordID, cb_readdir,
                               ptr<readdir_state>, str, nfsstat3);
  
  void setattr_got_attr (sessionp s, chordID obj, ptr<fattr3> fa,
                         nfsstat3 status);
  void setattr_reply (sessionp s, ptr<fattr3> fa, bool ok);

  void getattr_found_inode (sessionp, chordID, recordp, ptr<cattr>, 
                            nfsstat3);
  void access_found_inode (sessionp, chordID, recordp, ptr<cattr>,
                           nfsstat3);
  uint32 access_check (inode_log_record*, uint32 access_req);
  
  void lookup_found_link (sessionp, recordp, nfsstat3);
  void lookup_found_link_dup (sessionp, int, recordp, nfsstat3);
  void cancel_zombie_intend (sessionp, int phase, chordID dir, str name, bool);
  bool check_exclusive_link (chordID, str);
  void clear_exclusive_link (sessionp);
  void exclusive_phase_two (sessionp, chordID, str, bool);
  void create_found_inode (sessionp, chordID);
  void check_dir_exists (sessionp);

  void do_mkdir (sessionp);
  void link_found_inode (sessionp, chordID, str, recordp,
                         ptr<cattr>, nfsstat3);
  void link_inode_found_dir (sessionp, chordID, str,
                             recordp, ptr<cattr>, nfsstat3);
  void link_inode_to_file (sessionp, recordp, chordID, str);
  void link_inode_to_file_done (sessionp, recordp, chordID, bool);
  void link_inode_to_dir_done (sessionp, recordp, chordID, bool);
  
  void do_symlink (sessionp);
  void symlink_done (sessionp, chordID, recordp, bool);
  
  bool do_rmdir (sessionp, recordp, ptr<readdir_state>, str, nfsstat3);
  void rmdir_found_link (sessionp, chordID, recordp, nfsstat3);
  void remove_found_link (sessionp, int, recordp, nfsstat3);
  void remove_reply (sessionp, recordp, bool);

  void write_done (sessionp, uint32, bool);

  bool readdir_reply (sessionp, chordID, ptr<readdir_state>, str, nfsstat3);
  void readdir_found_inode (sessionp, chordID, recordp, ptr<cattr>, nfsstat3);

  void rename_found_link_old (sessionp, chordID, recordp, nfsstat3);
  void rename_found_link_new (sessionp, recordp, recordp, nfsstat3);
  void rename_done (sessionp, recordp, recordp, bool);

  void read_done (sessionp, ptr<read_state>, nfsstat3);
  void readlink_got_data (sessionp, ptr<read_state>, nfsstat3);

  void find_dir_parent_done (sessionp s, chordID p, nfsstat3 status);
  void find_dir_parent_check_snapshot (sessionp s, chordID dir, cb_chordid cb,
                                       chordID p, nfsstat3 status);
  void find_dir_parent (sessionp s, chordID dir, cb_chordid cb);

  void build_snapshot_with_mv (sfs_aid aid, ptr<materialized_view> mv);
  void load_snapshot (nfscall *sbp, ptr<materialized_view> mv,
                      chordID head, chordID loc, ptr<snapshot> s);
  void snapshot_created (ptr<log>, chordID, ptr<snapshot>);

public:
  cfsrw (ref<view> v, chordID vid, bigint& verf,
         callback<void, nfscall *, void *, nfs_fh3>::ref r);

  bool uid (nfscall *, chordID &);
  bool readonly (nfscall *);
  void dispatch (nfscall *, ptr<materialized_view>, bool need_snapshot = true);
  sessionp get_session (nfscall *sbp, ptr<materialized_view> mv,
                        bool need_snapshot);
 
  void readdir (sessionp, chordID dir, ptr<readdir_state>, cb_readdir cb);
  void lookup (sessionp, chordID, str, cb_lookup, bool check_intend = false);
  void read (sessionp, ptr<read_state>, cb_read cb);
  void write (sessionp, chordID, uint64, uint32, const char*, cb_bool);
  void find_inode (sessionp, chordID, cb_inode);
  void getattr (sessionp, chordID, cb_fa);
  
  void build_snapshot (sfs_aid aid, const vervec &);
  void build_snapshot (sfs_aid aid);

  static unsigned blocksize;
};
  
#endif

