
#ifndef CFSRWCD_H
#define CFSRWCD_H

#include "arpc.h"
#include "sfsclient.h"
#include "sfsmisc.h"
#include "dhash_prot.h"
#include "dhash.h"

#include "vervec.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "cfsa_prot.h"
#include "scan.h"
#include "debug.h"

#include "attrcache.h"
#include "fcache.h"
#include "cfsrw.h"

// file/attr cache entry is considered recent if it is 120 seconds or
// less old
#define CACHE_RECENT 120    

typedef callback<void, nfs_fh3*>::ref cbfh_t;

class cfsrw;
class cfsrwcd
{
  sfsprog &_prog;
  chordID _vid;
  bigint _verf;
  writeverf3 _verf3;
  
  cfsrw_attr_cache _ac;
  lrucache<nfs_fh3, file_cache *> _fc;
  str _cdir;
  
  ref<view> _view;
  cfsrw _cfsrw;

  void dispatch_with_view (nfscall *, ptr<materialized_view>);

  void flushfattr (unsigned int proc, void *arg, void *res);
  void getfattr (unsigned int proc, sfs_aid aid, void *arg, void *res);
  void reply (nfscall *, void *, nfs_fh3);
  
  str  fh2fn(nfs_fh3 fh);

  void read_from_cache (nfscall *, file_cache *);
  void read_from_cache_open (nfscall *, file_cache *, ptr<aiofh>, int);
  void read_from_cache_read (nfscall *, file_cache *, ptr<aiobuf>,
                             ssize_t, int);

  void write_to_cache (nfscall *, file_cache *);
  void write_to_cache_open (nfscall *, file_cache *, ptr<aiofh>, int);
  void write_to_cache_write (nfscall *, file_cache *, ptr<aiobuf>,
                             ssize_t, int);

  void truncate_cache (nfscall *, file_cache *, uint64);
  void truncate_cache_open (nfscall *, file_cache *, uint64, ptr<aiofh>, int);
  void truncate_cache_truncate (nfscall *, file_cache *, uint64, int);

  void run_rpcs (file_cache *);
  bool dont_run_rpc (nfscall *nc);
  
  bool fcache_dispatch (nfscall *);
  void fcache_check_reply (nfscall *sbp, void *res);

  void flush_cache (file_cache *e, nfscall *sbp);
  void flush_cache_with_mv (file_cache *, nfscall *, ptr<materialized_view>);
  void flush_done (file_cache *e, nfscall *sbp, bool ok);

  void fill_cache (file_cache *, callback<void,bool,bool>::ref, nfscall *);
  void fill_cache_with_mv (file_cache *, callback<void,bool,bool>::ref,
                           nfscall *, ptr<materialized_view>);
  void fill_done (file_cache *e, bool done, bool ok);

  void file_cache_gc_remove_cb (int);

  static void file_closed (int) {}

public:
  static aiod* a;

  void got_view (callback<void, cfsrwcd*, nfs_fh3*>::ref, ptr<view>);
  void dispatch (ref<nfsserv>, nfscall*);
  void dispatch_nfs (nfscall*);

  cfsrwcd (sfsprog &p, chordID view,
           callback<void, cfsrwcd*, nfs_fh3*>::ref cb);
  
  void file_cache_insert (nfs_fh3 fh) {
    file_cache *f = New file_cache(fh);
    _fc.insert(fh,f);
  } 
    
  file_cache *file_cache_lookup (nfs_fh3 fh) {
    file_cache **e = _fc[fh];
    if (e) {
      assert(*e);
      return *e;
    }
    return 0;
  }
  
  void file_cache_gc_remove (file_cache *e);
};

#endif

