
#ifndef CFSRW_CACHE_H
#define CFSRW_CACHE_H

#include "async.h"
#include "sfscrypt.h"
#include "dhash.h"
#include "vec.h"
#include "lrucache.h"
#include "record.h"
#include "cfsro.h"

//
// see cache.C on caching policy
//

extern ptr<sfsro_data> sfsro_decode (ptr <dhash_block> blk);

struct cached_block {
  enum cached_type {
    DHASH = 0,
    LOG,
    PKEY,
    SFSRO
  };
  unsigned short type;
  unsigned short touches;
  ptr<dhash_block> block;
  ptr<log_record> record;
  ptr<sfsro_data> sfsro;

  cached_block () { touches = 0; }
};

typedef callback<void, ptr<sfsro_data> >::ref cbsfsro;

class cfsrw_dhashclt
{
  static dhashclient *dhc;
  static lrucache<chordID, cached_block, hashID> record_cache;
  static lrucache<chordID, cached_block, hashID> sfsro_cache;
  static lrucache<chordID, cached_block, hashID> block_cache;

  static void copy_sfsro (cbsfsro cb, ptr<sfsro_data>);
  static void return_sfsro (cbsfsro cb, ptr<sfsro_data>);
  static void copy_record (cb_recordp cb, ptr<log_record>);
  static void return_record (cb_recordp cb, ptr<log_record>);
 
  static void retrieve_cb
    (cbretrieve_t cb, bool insert, chordID id, ptr<dhash_block> b);
  static void retrieve_record_cb
    (int tries, cb_recordp cb, bool insert, chordID id, ptr<dhash_block> b);
  static void retrieve_pk_cb (cb_recordp cb, ptr<dhash_block> b);
  static void retrieve_pk_fast_cb
    (chordID id, cb_recordp cb, ptr<dhash_block> b);
  static void retrieve_sfsro_cb
    (cbsfsro cb, bool insert, chordID id, ptr<dhash_block> b);
  static void retrieve_sfsro_readonly_cb
    (cbsfsro cb, bool insert, chordID id, ptr<dhash_block> b);

  static int record_insert;
  static int record_fetch;
  static int cached_record_fetch;
  static int sfsro_insert;
  static int sfsro_fetch;
  static int cached_sfsro_fetch;
  static int ch_insert;
  static int pk_insert;
  static int pk_fetch;
 
  static void insert_ok_cb (cbinsert_t cb, chordID);
  static void insert_pk_cb (recordp, cbinsert_t cb,
                            bool, ptr<insert_info>);
  static void insert_pk_fast_cb (chordID hash, sfs_pubkey2 k, sfs_sig2 s,
				 recordp e, cbinsert_t cb,
				 bool, ptr<insert_info>);
  static void insert_record_cb (int, recordp, cbinsert_t cb,
                                bool, ptr<insert_info>);
  static void insert_sfsro_cb (ptr<sfsro_data>, cbinsert_t cb,
                               bool, ptr<insert_info>);
  static void insert_ch_cb (ptr<dhash_block>, cbinsert_t cb,
                            bool, ptr<insert_info>);

public:
  static void init();
  static void stat();

  static void retrieve_ch (chordID id, cbretrieve_t cb);
  static void retrieve_pk (chordID id, cb_recordp cb);
  static void retrieve_record (chordID id, cb_recordp cb);
  static void retrieve_sfsro (chordID id, cbsfsro cb);
  static void retrieve_sfsro_readonly (chordID id, cbsfsro cb);

  static void insert_pk (ptr<sfspriv>, const char *, size_t,
                         cbinsert_t, recordp r = 0);
  static void insert_pk (sfs_pubkey2, sfs_sig2, const char *, size_t,
                         cbinsert_t, recordp r = 0);
  static void insert_pk (chordID, sfs_pubkey2, sfs_sig2,
                         const char *, size_t, cbinsert_t, recordp r = 0);
  static void insert_ch (const char *, size_t, cbinsert_t cb);
  static void insert_ch (chordID, const char *, size_t, cbinsert_t cb);
  static void insert_record (const char *, size_t, cbinsert_t cb, recordp r=0);
  static void insert_record (chordID, const char *, size_t,
                             cbinsert_t, recordp r=0);
  static void insert_sfsro (const char *, size_t,
                            cbinsert_t cb, ptr<sfsro_data> s = 0);
  static void insert_sfsro (chordID, const char *, size_t,
                            cbinsert_t cb, ptr<sfsro_data> s = 0);

  static bool has_sfsro (chordID);
  static bool has_block (chordID);
  static bool has_record (chordID);
};

#endif

