
#ifndef CFSRW_LOG_H
#define CFSRW_LOG_H

#include "cfsa_prot.h"
#include "record.h"

class cfsa_signhead_res;
class log : public virtual refcount 
{
private:
  chordID _head;
  sfs_aid _writer;
  bool _readonly;
  chordID _root;

  bool _refreshed;
  bool _loading;
  vec<cb_bool> _appends_cb;
  vec<recordp> _appends_e;
  chordID _snapshot;
  bool _has_snapshot;
  bool _snapshot_new;
  ptr<log_record> _head_record;

  int _stores;
  int _failed_stores;
  vec<cb_bool> _store_cbs;

  void sign_block_cb
    (callback<void, sfs_pubkey2, sfs_sig2, bool>::ref,
     cfsa_signhead_res *, clnt_stat);
  void sign (const str&, callback<void, sfs_pubkey2, sfs_sig2, bool>::ref);

  void store_head_after_append(bool);
  void head_stored(ptr<vec<cb_bool> >, ptr<head_log_record>, bool);
  void do_append(recordp head);
  
  void store_record(chordID, const str&, cb_bool, recordp e);
  void store_record_cb(cb_bool, bool, chordID);
  void get_record_cb(cb_recordp, int, str);

  void append_log_cb1(recordp, cb_bool, recordp);
  void append_log_cb2(chordID, cb_bool, recordp, bool);

  void load_head(recordp);
  void load_head_cb(cb_recordp, recordp);
  void store_head(recordp, const str, cb_bool, sfs_pubkey2, sfs_sig2, bool);
  
  static void dhash2log(cb_recordp, chordID, ptr<dhash_block>);
  static qhash<sfs_aid, ref<aclnt> > cahash;

public:
  log(chordID);
  log(recordp);
  log(chordID, cb_recordp);

  chordID head () const { return _head; }
  chordID root () const { return _root; }
  void writer (sfs_aid a) { _writer = a; }
  sfs_aid writer () const { return _writer; }

  bool readonly () const { return _readonly; }
  
  void append(const vervec &v, recordp, cb_bool);
  void add_snapshot(chordID s);
  bool has_snapshot() const { return _has_snapshot; }
  chordID snapshot() const { return _snapshot; }

  static void locate(chordID, bool keyhashed, cb_recordp);
  static void prev_record(recordp, cb_recordp);
  static ptr<aclnt> ca(sfs_aid aid);
};

#endif

