#include "ranges.h"

ranges::ranges(uint64 size)
  : _size(size)
{
  _l.insert_head(New range(0, 0));
}

ranges::~ranges()
{
  range *p, *np;
  for(p = _l.first; p; p = np){
    np = (p->_link).next;
    delete p;
  }
}

void
ranges::add(uint64 start, uint64 len)
{
  // Look at each gap. _l guaranteed to start w/ an entry at zero.
  range *p, *np;
  for(p = _l.first; p; p = np){
    np = (p->_link).next;
    assert(np == 0 || np->_start >= p->_start + p->_len);
    uint64 gs = p->_start + p->_len;
    uint64 ge = (np ? np->_start : _size);
    if(start+len >= gs && start < ge && gs != ge){
      uint64 ns = max(gs, start);
      uint64 ne = min(ge, start+len);
      range *r = New range(ns, ne - ns);
      r->_link.next = np;
      r->_link.pprev = &(p->_link.next);
      p->_link.next = r;
      if(np)
        np->_link.pprev = &(r->_link.next);
    }
  }
}

// Is the entire start/len interval filled in?
bool
ranges::check(uint64 start, uint64 len)
{
  // Does any gap overlap with start/end?
  range *p, *np;
  for(p = _l.first; p; p = np){
    np = (p->_link).next;
    uint64 gs = p->_start + p->_len;
    uint64 ge = (np ? np->_start : _size);
    if(start+len > gs && start < ge && gs != ge){
      return false;
    }
  }
  return true;
}

bool
ranges::gaps(callback<void, uint64, uint64>::ref cb)
{
  range *p, *np;

  for(p = _l.first; p; p = np){
    np = (p->_link).next;
    uint64 gs = p->_start + p->_len;
    uint64 ge = (np ? np->_start : _size);
    if(gs != ge)
      cb(gs, ge - gs);
  }

  // XXX: ???
  return true;
}

uint64
ranges::end()
{
  uint64 e = 0;
  range *p, *np;
  for(p = _l.first; p; p = np){
    np = (p->_link).next;
    e = p->_start + p->_len;
  }
  return e;
}

// is the entire start/len interval filled in?
bool
ranges::filled(uint64 start, uint64 len) const
{
  // does any gap overlap with start/end?
  range *p, *np;
  for (p = _l.first; p; p = np) {
    np = (p->_link).next;
    uint64 gs = p->_start + p->_len;
    uint64 ge = (np ? np->_start : _size);
    if (start+len > gs && start < ge && gs != ge)
      return false;
  }
  return true;
}

bool
ranges::has_next_gap(uint64 off, uint64 &start, uint64 &size) const
{
  range *p, *np;

  for (p = _l.first; p; p = np) {
    np = (p->_link).next;
    uint64 gs = p->_start + p->_len;
    uint64 ge = (np ? np->_start : _size);
    if (gs != ge && gs >= off) {
      start = gs;
      size = ge - gs;
      return true;
    }
  }
  return false;
}

