#ifndef CFSRW_SCAN_H
#define CFSRW_SCAN_H

#include "nfs3_prot.h" // nfsstat3
#include "record.h"
#include "chord.h"
#include "ranges.h"
#include "logiter.h"

class readdir_state;
class read_state;

typedef callback<void, ptr<read_state>, nfsstat3>::ptr cb_read;
typedef callback<void, recordp, nfsstat3>::ref cb_lookup;
typedef callback<void, chordID, nfsstat3>::ref cb_chordid;
typedef callback<void, recordp, ptr<cattr>, nfsstat3>::ref cb_inode;
typedef callback<void, ptr<fattr3>, nfsstat3>::ref cb_fa;
typedef callback<bool, ptr<readdir_state>, str, nfsstat3>::ref cb_readdir;

struct read_state {
public:
  chordID inum;
  uint64 offset;
  uint32 count;
  uint64 size;
  bool valid_size;
  ranges *rng;
  char *buf;
  uint32 rcount;
  bool eof;
  read_state(chordID i, uint64 o, uint32 c)
    : inum(i), offset(o), count(c), size(0), valid_size(false),
      rcount(0), eof(false)
  {
    if (count > 8192) count = 8192;
    rng = New ranges;
    buf = New char[count];
  }
  ~read_state() { if (buf) delete buf; if (rng) delete rng; }
};

class lookup_state {
public:
  chordID dir;
  str name;
  cb_lookup cb;
  ptr<vec<chordID> > ci;
  lookup_state (chordID d, str n, cb_lookup c, ptr<vec<chordID> > v)
    : dir(d), name(n), cb(c), ci(v) {}
  ~lookup_state () {}
};

class readdir_state {
public:
  readdir_state(readdir3args *arg)
    : nfsres(NFS3_OK)
  {
    nfsres.resok->reply.eof = false;
    direntp = &(nfsres.resok->reply.entries);
    cookie = arg->cookie;
    count = arg->count - 184;
    n = 0;
  }
  readdir_state()
    : nfsres(NFS3_OK)
  {
    nfsres.resok->reply.eof = false;
    direntp = &(nfsres.resok->reply.entries);
    cookie = 0;
    count = 0;
    n = 0;
  }
  ~readdir_state() {}

  unsigned int count;
  uint64 cookie;
  unsigned int n;
  readdir3res nfsres;
  rpc_ptr<entry3> *direntp;
  vec<recordp> changed;
};

class logscan : public virtual refcount {
private:
  static void scan_log (cbliter cb, ptr<materialized_view> mv);
  static void scan_log (cbliter cb, ptr<materialized_view> mv, vervec stop);
  
  static void find_inode_log_iterator
              (chordID, cb_inode, ptr<cattr>, iterinfo*);
  static void lookup_log_iterator (ref<lookup_state> lu, iterinfo*);
  static void find_dir_parent_log_iterator (chordID, cb_chordid, iterinfo*);

  static void scan_dir_log_iterator
              (chordID dir, ptr<readdir_state>, cb_readdir cb, iterinfo*);
  static bool readdir_check_dir_entry
              (ptr<readdir_state>, chordID dir, cb_readdir cb,
	       str name, chordID inum, bool finished);

  static void sfrxx (ptr<read_state>, write_log_record *, uint64, uint64);
  static void do_read_got_chunk (ptr<read_state>, cb_read, iterinfo*);

public:
  static void readdir
              (ptr<materialized_view> mv, vervec stop,
	       chordID dir, ptr<readdir_state>, cb_readdir);
  static void find_inode
              (ptr<materialized_view> mv, vervec stop,
	       chordID inum, ptr<cattr>, cb_inode cb);
  static void lookup
              (ptr<materialized_view> mv, vervec stop,
	       chordID dir, str name, cb_lookup cb, bool check_intend = false);
  static void find_dir_parent
              (ptr<materialized_view> mv, vervec stop,
	       chordID inum, cb_chordid cb);
  static void read
              (ptr<materialized_view> mv, vervec stop,
	       ptr<read_state>, cb_read cb);
};

#endif

