
#include "cache.h"
#include "dhash.h"
#include "record.h"
#include "log.h"
#include "view.h"

view::view(callback<void, ptr<view> >::ref cb, chordID v)
{
  cfsrw_dhashclt::retrieve_ch(v, wrap(mkref(this), &view::got_view, cb));
}

void
view::got_view(callback<void, ptr<view> >::ref cb, ptr<dhash_block> blk)
{
  if (blk == 0) {
    warn << "could not fetch view block\n";
    cb((ptr<view>)0);
    return;
  }

  str ins(blk->data, blk->len);
  xin *in = New xin(ins);
  parse(in);
  delete in;
  cb(mkref(this));
}

chordID
view::store(callback<void, bool, chordID>::ref cb)
{
  str data = serialize();
  chordID id = compute_hash(data.cstr(), data.len());
  cfsrw_dhashclt::insert_ch(data.cstr(), data.len(), cb);
  return id;
}

void
view::sort()
  // slow but simple sort
{
  vec<chordID> h;
  vec<chordID> l;
  vec<bool> done;
  bool found = true;
  for (unsigned i=0; i<_head.size(); i++)
    done.push_back(false);

  while (found) {
    found = false;
    chordID sv;
    unsigned s = 0;
    for (unsigned i=0; i<_head.size(); i++) {
      if (!done[i]) {
	if (!found || _head[i] < sv) {
          found = true;
	  sv = _head[i];
	  s = i;
	}
      }
    }
    if (found) {
      h.push_back(_head[s]);
      l.push_back(_latest[s]);
      done[s] = true;
    }
  }

  _head = h;
  _latest = l;
}

void
view::load_record(ptr<materialized_view> mv,
                  callback<void, ptr<materialized_view> >::ref cb, recordp e)
{
  if (!e) {
    warn << "cannot load record to materialize view\n";
    cb((ptr<materialized_view>)0);
  }
  else if (e->template record<head_log_record>()) {
    mv->add_head_record(e);
    log::locate
      (e->prevID(), false, wrap(mkref(this), &view::load_record, mv, cb));
  }
  else {
    mv->add_record(e);
    if (mv->finished()) {
      mv->make_version();
      cb(mv);
    }
  }
}

void
view::materialize(callback<void, ptr<materialized_view> >::ref cb)
{
  ptr<materialized_view> mv = New refcounted<materialized_view>();
  mv->_root = _root;
  for (unsigned i=0; i<_head.size(); i++) {
    mv->_head.push_back(_head[i]);
    mv->_head_record.push_back((ptr<log_record>)0);
    mv->_latest.push_back((ptr<log_record>)0);
    log::locate
      (_latest[i], (_latest[i]==_head[i]),
       wrap(mkref(this), &view::load_record, mv, cb));
  }
}


