#include "snapshot.h"

void
init_attr (sfsro_inode_reg *inode)
{
  inode->mtime.seconds = 0;
  inode->mtime.nseconds = 0;
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.uid = 0;
  inode->attr.gid = 0;
  inode->attr.rdev.major = 0;
  inode->attr.rdev.minor = 0;
}

void
init_attr (sfsro_inode_lnk *inode)
{
  inode->mtime.seconds = 0;
  inode->mtime.nseconds = 0;
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.uid = 0;
  inode->attr.gid = 0;
  inode->attr.rdev.major = 0;
  inode->attr.rdev.minor = 0;
}

void
set_attr (sfsro_inode_reg *inode, cattr *ca)
{
  if (ca->valid_mtime) inode->mtime.seconds = ca->mtime.seconds;
  inode->mtime.nseconds = 0;
  if (ca->valid_size)
    inode->attr.size = ca->size;
  if (ca->valid_mode) inode->attr.mode = ca->mode;
  if (ca->valid_uid) inode->attr.uid = ca->uid;
  if (ca->valid_gid) inode->attr.gid = ca->gid;
  if (ca->valid_rdev) inode->attr.rdev.major = ca->rdev.major;
  if (ca->valid_rdev) inode->attr.rdev.minor = ca->rdev.minor;
}

void
set_attr (sfsro_inode_lnk *inode, cattr *ca)
{
  if (ca->valid_mtime) inode->mtime.seconds = ca->mtime.seconds;
  inode->mtime.nseconds = 0;
  if (ca->valid_size) inode->attr.size = ca->size; // XXX what do i do?
  if (ca->valid_mode) inode->attr.mode = ca->mode;
  if (ca->valid_uid) inode->attr.uid = ca->uid;
  if (ca->valid_gid) inode->attr.gid = ca->gid;
  if (ca->valid_rdev) inode->attr.rdev.major = ca->rdev.major;
  if (ca->valid_rdev) inode->attr.rdev.minor = ca->rdev.minor;
}

void 
get_random_ID (chordID *ID, uint len) 
{
  char xx[len];
  rnd.getbytes (xx, len);
  mpz_set_rawmag_be (ID, xx, len);
}

str
get_dirname (nfspath3 path)
{
  unsigned short *a;
  a = (unsigned short*) path.cstr();
  unsigned short d = (*a) % 240;
  return strbuf ("/var/tmp/ivy/scache/%d", d);
}

str 
get_filename (nfspath3 path)
{
  unsigned short *a;
  a = (unsigned short*) path.cstr();
  unsigned short d = (*a) % 240;
  return strbuf ("/var/tmp/ivy/scache/%d/%s", d, path.cstr());
}

void 
set_cfsrwattr (cattr *ca, sfsro_inode_reg *rnode)
{
  ca->valid_size = (rnode->attr.size == 666 ? 0 : 1);
  ca->size = rnode->attr.size;
  ca->valid_mode = (rnode->attr.mode == 666 ? 0 : 1);
  ca->size = rnode->attr.mode;
  ca->valid_uid = (rnode->attr.uid == 666 ? 0 : 1);
  ca->size = rnode->attr.uid;
  ca->valid_gid = (rnode->attr.gid == 666 ? 0 : 1);
  ca->size = rnode->attr.gid;
  ca->valid_rdev = (rnode->attr.rdev.major == 666 ? 0 : 1);
  ca->rdev.major = rnode->attr.rdev.major;  
  ca->rdev.minor = rnode->attr.rdev.minor;  
}

void
combine_attr (ptr<cattr> ca, ptr<sfsro_data> sdata)
{
  sfsro_inode *inode = sdata->inode;

  if (!ca->truncated) {
    if (inode->type == SFSROLNK && ca->size < inode->lnk->attr.size)
      ca->size = inode->lnk->attr.size;
    else if (inode->type != SFSROLNK && ca->size < inode->reg->attr.size)
      ca->size = inode->reg->attr.size;
  }
  if (inode->type == SFSROLNK) {
    sfsro_inode_lnk *lnode= inode->lnk;
    ca->nlink += lnode->nlink;
    ca->mode = ca->valid_mode ? ca->mode : lnode->attr.mode;
    ca->uid = ca->valid_uid ? ca->uid : lnode->attr.uid;
    ca->gid = ca->valid_gid ? ca->gid : lnode->attr.gid;
    if (!ca->valid_mtime || ca->mtime < lnode->mtime) {
      ca->mtime = lnode->mtime;
      ca->valid_mtime = true;
    }
    ca->rdev.major = ca->valid_rdev ? ca->rdev.major : lnode->attr.rdev.major;
    ca->rdev.minor = ca->valid_rdev ? ca->rdev.minor : lnode->attr.rdev.minor;
    if(ca->ctime < lnode->ctime)
      ca->ctime = lnode->ctime;
  }
  else {
    sfsro_inode_reg *rnode= inode->reg;
    ca->nlink += rnode->nlink;
    ca->mode = ca->valid_mode ? ca->mode : rnode->attr.mode;
    ca->uid = ca->valid_uid ? ca->uid : rnode->attr.uid;
    ca->gid = ca->valid_gid ? ca->gid : rnode->attr.gid;
    if (!ca->valid_mtime || ca->mtime < rnode->mtime) {
      ca->mtime = rnode->mtime;
      ca->valid_mtime = true;
    }
    ca->rdev.major = ca->valid_rdev ? ca->rdev.major : rnode->attr.rdev.major;
    ca->rdev.minor = ca->valid_rdev ? ca->rdev.minor : rnode->attr.rdev.minor;
    if(ca->ctime < rnode->ctime)
      ca->ctime = rnode->ctime;
  }
  ca->valid_size = true;
  ca->valid_mode = true;
  ca->valid_uid = true;
  ca->valid_gid = true;
  ca->valid_mtime = true;
  ca->valid_rdev = true;
}

ptr<log_record>
set_log_record (bool is_lookup, ptr<sfsro_data> sdata, chordID dir, str name)
{  
  chordID inum;
  sfsro_inode *inode = sdata->inode;
  cattr ca;

  int itype;
  if (inode->type == SFSROLNK) {
    str dearm = dearmor64A ((*(*inode).lnk).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    itype = inode_log_record::T_SYMLINK;
    set_cfsrwattr (&ca, inode->lnk); // this is not used?
  } else {
    str dearm = dearmor64A ((*(*inode).reg).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    if (inode->type == SFSROREG || inode->type == SFSROREG_EXEC)
      itype = inode_log_record::T_FILE;
    else itype = inode_log_record::T_DIR;
    set_cfsrwattr (&ca, inode->reg); // this is not used?
  }

  ptr<log_record> le;
  if (is_lookup) {
    int itype = inode_log_record::T_FILE;
    if (inode->type == SFSROLNK)
      itype = inode_log_record::T_SYMLINK;
    else if (inode->type == SFSRODIR)
      itype = inode_log_record::T_DIR;
    le = New refcounted <link_log_record> (inum, itype, dir, name);
  }
  else
    le = New refcounted <inode_log_record> (itype, inum, ca);

  return le;
}

bool 
dup_entry (rpc_ptr<entry3> *dep, sfsro_dirent *e)
{
  if (!dep) 
    return false;
  rpc_ptr<entry3> de = *dep;
  while (de && de->name != e->name)
    de = de->nextentry;
  if (de)
    return true;
  return false;
}

ref<dbrec>
id2dbrec(chordID id)
{
  str whipme = id.getraw ();
  void *key = (void *)whipme.cstr ();
  int len = whipme.len ();

  ptr<dbrec> q = New refcounted<dbrec> (key, len);
  return q;
}

chordID
dbrec2id (ptr<dbrec> r)
{
  str raw = str ( (char *)r->value, r->len);
  chordID ret;
  ret.setraw (raw);
  return ret;
}

int
dbcompare (ref<dbrec> a, ref<dbrec> b)
{
  chordID ax = dbrec2id (a);
  chordID bx = dbrec2id (b);
  if (ax < bx)
    return -1;
  else if (ax == bx)
    return 0;
  else
    return 1;
}

