#!/usr/bin/perl -w
#
# Executes all test_*.pl files in a freshly build cfsrw directory.
#
# Test files may have a ``# ARGS: [BLAH]'' line, where BLAH is the arguments
# you would like to pass to that test.
#

use strict;
$| = 1;

my $SETUP = "../script/setup.pl";
my $SFS_PREFIX = "/usr/local/";
my $CHORD_PREFIX = "/usr/local/";
my $CFSRW_PREFIX = "/usr/local/";
my $CFSRW = "~/.cfsrw";

my $testdir; # e.g. ~/cfsrw/test
my $fsdir;   # .e.g /cfsrw/chord:xxx

sub prepare {
    if(!defined($fsdir)){
        # create fresh CFSRW directory
        my $z = `$SETUP +i -s --sfsprefix $SFS_PREFIX --chordprefix $CHORD_PREFIX --cfsrwprefix $CFSRW_PREFIX --cfsrw $CFSRW | tail -1`;
        if($z =~ /^cd (.*chord:[^\n]+)/){
            $fsdir = $1;
        } else {
            print STDERR "alltests.pl: setup.pl failed\n";
            exit(1);
        }
    }
}

#
# Use setup script to create a fresh CFSRW directory
sub main {
    $testdir = `pwd`;
    $testdir =~ s/\n.*//;
    process_args(@ARGV);
    prepare();

    print STDERR "testdir $testdir\n";
    print STDERR "fsdir $fsdir\n";

  # XXX: should happen on build!
  system "cd $testdir/tests && make > /dev/null";

  # which test files?
  my $ret = 0;
  my $o = `find $testdir/tests -type f -perm +0111 -print`;
  my @t = split /\n/, $o;

  foreach my $t (@t) {
    -d $t and next;
    -x $t or warn "Can't execute $t. Skipping.\n" and next;

    my $type = `file $t`;
    my $a;
    if ($type =~ m/script/) {
      $a = `egrep "^#[ ]*ARGS:" $t`;
      $a =~ s/^#\s+ARGS:\s+(.*)/$1/;
    } elsif ($type =~ m/executable/) {
      $a = `egrep "[:space:\/\*]*ARGS:" $t.c`;
      $a =~ s/^[\s\/\*]*ARGS:\s+(.*)/$1/;
    }
    chomp $a;

    print "* $t $a ";
    if(system "cd $fsdir && $t $a") {
      print "[ FAILED ]\n";
      $ret++;
    } else {
      print "[ OK ]\n";
    }
  }

  exit $ret;
}

#
# Processes arguments and sets options.
#
sub process_args {
  my @args = @_;

  while(@args) {
    $_ = shift @args;
    if(/^--cfsrw$/)           { $CFSRW = shift @args; next; }
    if(/^--sfsprefix$/)       { $SFS_PREFIX = shift @args; next; }
    if(/^--chordprefix$/)     { $CHORD_PREFIX = shift @args; next; }
    if(/^--cfsrwprefix$/)     { $CFSRW_PREFIX = shift @args; next; }
    if(/^--setup$/)           { $SETUP = shift @args; next; }
    if(/^--dir$/)             { $fsdir = shift @args; next; }

    print STDERR "Unrecognized option $_\n";;
    usage();
  }
}

sub usage {
    print STDERR "Usage: alltests [options...]\n";
    print STDERR "  --dir /cfsrw/xxx (no setup, use existing dir)\n";
    print STDERR "  --cfsrw dir (default ~/.cfsrw)\n";
    print STDERR "  --sfsprefix dir (default /usr/local)\n";
    print STDERR "  --chordprefix dir (default /usr/local)\n";
    print STDERR "  --cfsrwprefix dir (default /usr/local)\n";
    print STDERR "  --setup file (default ../script/setup.pl)\n";
    exit(1);
}

main();
