#!/usr/bin/perl -w
# ARGS: +v

# Create / rename / delete test.
# You can give it multiple equivalent directories.

use strict;

my @dirs; # A set of equivalent directory names.
my $maxiters = 200; # -n
my $maxfiles = 10;
my $verbose = 1;

&process_args(@ARGV);
@ARGV = ();

if($#dirs == -1){
    $dirs[0] = ".";
}

my $prefix = "tf" . $$ . "-";

my %state;

my $iter;
for($iter = 0; $iter < $maxiters; $iter++){
    my $fn1 = $prefix . int(rand($maxfiles));
    my $fn2 = $prefix . int(rand($maxfiles));
    my $d = rchoice(@dirs);
    my $f1 = $d . "/" . $fn1;
    my $f2 = $d . "/" . $fn2;
    my $junk = rand();

    my $choice = rand();
    if($choice < 0.25){
        # create
        $verbose >= 1 and print "create $f1\n";
        if(!open(F, ">$f1")){
            $verbose >= 1 and print "crdtest.pl: open/create $f1: $!\n";
            cleanup(1);
        }
        print F $junk;
        close(F);
        $state{$fn1} = $junk;
    } elsif($choice < 0.5){
        # rename
        my $st = rename($f1, $f2);
        $verbose >= 1 and print "rename $f1 $f2 -> $st\n";
        if(defined($state{$fn1}) && !$st){
            # rename should have succeeded
            $verbose >= 1 and print "crdtest.pl: rename: $!\n";
            cleanup(1);
        }
        if(!defined($state{$fn1}) && $st){
            # rename should not have succeeded
            $verbose >= 1 and print "crdtest.pl: rename worked, oops\n";
            cleanup(1);
        }
        if($st && ($f1 ne $f2)){
            $state{$fn2} = $state{$fn1};
            delete $state{$fn1};
        }
    } elsif($choice < 0.75){
        # delete
        my $st = unlink($f1);
        $verbose >= 1 and print "delete $f1 -> $st\n";
        if(defined($state{$fn1}) && !$st){
            # unlink should have succeeded
            $verbose >= 1 and print "crdtest.pl: unlink: $!\n";
            cleanup(1);
        }
        if(!defined($state{$fn1}) && $st){
            # unlink should not have succeeded
            $verbose >= 1 and print "crdtest.pl: unlink worked, oops\n";
            cleanup(1);
        }
        if($st){
            delete $state{$fn1};
        }
    } elsif($choice < 0.9){
        # check
        my $st = open(F, $f1);
        $verbose >= 1 and printf("check $f1 -> %d\n", defined($st) ? $st : 0);
        if($st){
            if(defined($state{$fn1})){
                my $x = <F>;
                if($x ne $state{$fn1}){
                    $verbose >= 1 and print "crdtest.pl: read the wrong thing\n";
                    cleanup(1);
                }
            } else {
                $verbose >= 1 and print "crdtest.pl: check worked, oops\n";
                cleanup(1);
            }
            close(F);
        } else {
            if(defined($state{$fn1})){
                $verbose >= 1 and print "crdtest.pl: check failed, no file\n";
                cleanup(1);
            }
        }
    } else {
        # readdir
        $verbose >= 1 and print "readdir $d\n";
        if(!opendir(D, $d)){
            $verbose >= 1 and print "crdtest.pl: opendir: $!\n";
            cleanup(1);
        }
        my %xx;
        my $f;
        while($f = readdir(D)){
            if(substr($f, 0, length($prefix)) eq $prefix){
                $xx{$f} = 1;
            }
        }
        closedir(D);
        my @a = sort { $a cmp $b } keys(%xx);
        my @b = sort { $a cmp $b } keys(%state);
        if($#a != $#b){
            $verbose >= 1 and print "crdtest.pl: readdir: wrong number of files\n";
            cleanup(1);
        }
        my $i;
        for($i = 0; $i <= $#a; $i++){
            if($a[$i] ne $b[$i]){
                $verbose >= 1 and print "crdtest.pl: readdir oops\n";
                cleanup(1);
            }
        }
    }
}

cleanup(0);

sub rchoice {
    my @l = @_;
    my $i = int(rand($#l + 1));
    return $l[$i];
}

sub cleanup {
    my($status) = @_;
    my $i;
    for($i = 0; $i < $maxfiles; $i++){
        my $fn1 = $prefix . $i;
        my $f1 = $dirs[0] . "/" . $fn1;
        my $st = unlink($f1);
        if(defined($state{$fn1}) && !$st){
            # unlink should have succeeded
            $verbose >= 1 and print "crdtest.pl: unlink: $!\n";
            $status = 1;
        }
        if(!defined($state{$fn1}) && $st){
            # unlink should not have succeeded
            $verbose >= 1 and print "crdtest.pl: unlink worked, oops\n";
            $status = 1;
        }
    }
    exit($status);
}

#
# Processes arguments and sets options.
#
sub process_args {
  my @args = @_;

  while(@args) {
    $_ = shift @args;
    if(/-v/)                { $verbose++; next; }
    if(/\+v/)               { $verbose--; next; }
    if(/-n/)                { $maxiters = shift @args; next; }
    unless(/^-/)            { push @dirs, $_; next; }

    print "Usage: crdtest.pl [-n iters] directories...\n";
    exit(1);
  }
}
