/*
 * Test whether link counts are correctly maintained, in the
 * face of link(), unlink(), rename().
 *
 * ARGS: .
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <assert.h>

extern int errno;

void check_fds(int, int);
void check_links(char *file, int n);
void cleanup(int);

char d1[1024], d2[1024];
char f0[1024], f1[1024], f2[1024], f3[1024], f4[1024];

main(int argc, char *argv[])
{
  int pid = getpid();
  int fd;
  char *dir;

  if(argc != 2) {
    fprintf(stderr, "Usage: linktest directory-name\n");
    exit(1);
  }

  sprintf(d1, "%s/xda%d", argv[1], pid);
  sprintf(d2, "%s/xdb%d", argv[1], pid);
  sprintf(f0, "%s/f0", d1);
  sprintf(f1, "%s/f1", d1);
  sprintf(f2, "%s/f2", d1);
  sprintf(f3, "%s/f3", d2);
  sprintf(f4, "%s/f4", d2);

  if(mkdir(d1, 0777) != 0){
    fprintf(stderr, "linktest: mkdir %s: %s\n", d1, strerror(errno));
    cleanup(1);
  }

  if(mkdir(d2, 0777) != 0){
    fprintf(stderr, "linktest: mkdir %s: %s\n", d1, strerror(errno));
    cleanup(1);
  }

  fd = open(f0, O_CREAT|O_RDWR, 0666);
  if(fd < 0){
    fprintf(stderr, "linktest: creat %s: %s\n", f0, strerror(errno));
    cleanup(1);
  }
  close(fd);
  
  fd = open(f1, O_CREAT|O_RDWR, 0666);
  if(fd < 0){
    fprintf(stderr, "linktest: creat %s: %s\n", f1, strerror(errno));
    cleanup(1);
  }
  close(fd);
  
  check_links(f1, 1);

  assert(link(f1, f2) == 0);
  check_links(f1, 2);
  check_links(f2, 2);

  assert(rename(f0, f2) == 0);
  check_links(f1, 1);
  check_links(f2, 1);

  assert(unlink(f2) == 0);
  assert(link(f1, f2) == 0);
  check_links(f1, 2);
  check_links(f2, 2);

  assert(link(f1, f4) == 0);
  check_links(f1, 3);
  check_links(f2, 3);
  check_links(f4, 3);

  assert(rename(f2, f3) == 0);
  check_links(f1, 3);
  check_links(f3, 3);
  check_links(f4, 3);

  assert(rename(f1, f3) == 0);
  check_links(f3, 2);
  check_links(f4, 2);

  assert(unlink(f3) == 0);
  check_links(f4, 1);

  assert(link(f4, f1) == 0);
  assert(link(f1, f2) == 0);
  check_links(f4, 3);

  cleanup(0);
}

/*
 * Check the link count of a file.
 */
void
check_links(char *file, int n)
{
  int fd;
  struct stat st;

  fd = open(file, 0);
  if(fd < 0){
    fprintf(stderr, "linktest: open %s: %s\n", file, strerror(errno));
    cleanup(1);
  }
  if(fstat(fd, &st) < 0){
    fprintf(stderr, "linktest: fstat %s: %s\n", file, strerror(errno));
    cleanup(1);
  }
  close(fd);

  if(st.st_nlink != n){
    fprintf(stderr, "linktest: %s has %d links, wanted %d\n",
            file, st.st_nlink, n);
    cleanup(1);
  }
}

void
cleanup(int status)
{
  if(f1[0])
    unlink(f1);
  if(f2[0])
    unlink(f2);
  if(f3[0])
    unlink(f3);
  if(f4[0])
    unlink(f4);
  if(d1[0])
    rmdir(d1);
  if(d2[0])
    rmdir(d2);
  exit(status);
}
