/*
 * Test interaction between open file descriptors and link/rename.
 *
 * Open a file through two links, remove one, do both
 * file descriptors still work?
 *
 * Ought to do the link/rename via another mount of the file system.
 *
 * ARGS: .
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <assert.h>

extern int errno;

void check_fds(int, int);
void check_links(char *file, int n);
void cleanup(int);

char d1[1024], d2[1024];
char f1[1024], f2[1024], f3[1024], f4[1024];

main(int argc, char *argv[])
{
  int pid = getpid();
  int fd1, fd2, fd3, fd4;
  char *dir;

  if(argc != 2) {
    fprintf(stderr, "Usage: linktest directory-name\n");
    exit(1);
  }

  sprintf(d1, "%s/xda%d", argv[1], pid);
  sprintf(d2, "%s/xdb%d", argv[1], pid);
  sprintf(f1, "%s/f1", d1);
  sprintf(f2, "%s/f2", d1);
  sprintf(f3, "%s/f3", d2);
  sprintf(f4, "%s/f4", d2);

  if(mkdir(d1, 0777) != 0){
    fprintf(stderr, "linktest: mkdir %s: %s\n", d1, strerror(errno));
    cleanup(1);
  }

  if(mkdir(d2, 0777) != 0){
    fprintf(stderr, "linktest: mkdir %s: %s\n", d1, strerror(errno));
    cleanup(1);
  }

  fd1 = open(f1, O_CREAT|O_RDWR, 0666);
  if(fd1 < 0){
    fprintf(stderr, "linktest: creat %s: %s\n", f1, strerror(errno));
    cleanup(1);
  }
  
  check_links(f1, 1);

  if(link(f1, f2) < 0){
    fprintf(stderr, "linktest: link %s %s: %s\n",
            f1, f2, strerror(errno));
    cleanup(1);
  }
  check_links(f1, 2);
  check_links(f2, 2);
  assert(link(f1, f4) == 0);
  check_links(f2, 3);

  fd2 = open(f2, O_RDWR);
  if(fd2 < 0){
    fprintf(stderr, "linktest: open %s: %s\n", f2, strerror(errno));
    cleanup(1);
  }

  check_fds(fd1, fd2);
  assert(rename(f2, f3) == 0);
  check_links(f4, 3);
  check_fds(fd1, fd2);
  check_links(f3, 3);
  check_links(f4, 3);
  assert(rename(f1, f3) == 0);
  // check_links(f4, 2);
  // check_links(f3, 2);
  check_fds(fd1, fd2);
  
  assert((fd4 = open(f4, O_RDWR)) >= 0);
  check_fds(fd1, fd4);
  check_fds(fd2, fd4);

  cleanup(0);
}

/*
 * Check the link count of a file.
 */
void
check_links(char *file, int n)
{
  int fd;
  struct stat st;

  fd = open(file, 0);
  if(fd < 0){
    fprintf(stderr, "linktest: open %s: %s\n", file, strerror(errno));
    cleanup(1);
  }
  if(fstat(fd, &st) < 0){
    fprintf(stderr, "linktest: fstat %s: %s\n", file, strerror(errno));
    cleanup(1);
  }
  close(fd);

  if(st.st_nlink != n){
    fprintf(stderr, "linktest: %s has %d links, wanted %d\n",
            file, st.st_nlink, n);
    cleanup(1);
  }
}

/*
 * Make sure that the two file descriptors seem to point to
 * the same file.
 */
void
check_fds(int fd1, int fd2)
{
  char buf[512];

  assert(ftruncate(fd1, 0) == 0);
  assert(lseek(fd1, 0, 0) == 0);
  assert(write(fd1, "ab", 2) == 2);
  assert(fsync(fd1) == 0);

  assert(lseek(fd2, 0, 0) == 0);
  assert(read(fd2, buf, 512) == 2);
  assert(strncmp(buf, "ab", 2) == 0);

  assert(lseek(fd2, 1, 0) == 1);
  assert(write(fd2, "xy", 2) == 2);
  assert(fsync(fd2) == 0);

  assert(read(fd1, buf, 512) == 1);
  assert(strncmp(buf, "y", 1) == 0);
}

void
cleanup(int status)
{
  if(f1[0])
    unlink(f1);
  if(f2[0])
    unlink(f2);
  if(f3[0])
    unlink(f3);
  if(f4[0])
    unlink(f4);
  if(d1[0])
    rmdir(d1);
  if(d2[0])
    rmdir(d2);
  exit(status);
}
