/*
 * Simple program that writes to random places in a file
 * and checks that random reads return the right results.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>

void rwtest(char *filename);

main(int argc, char *argv[])
{
  int seed;
  char *fname;

  if(argc == 1){
    fname = malloc(256);
    sprintf(fname, "rwtest-%d", getpid());
  } else if(argc == 2){
    fname = argv[1];
  } else {
    fprintf(stderr, "Usage: rwtest [filename]\n");
    exit(1);
  }

  seed = time((time_t*) 0);
  seed <<= 16;
  seed |= (getpid() & 0xffff);
  srandom(seed);

  rwtest(fname);

  exit(0);
}

void
randomize(char *buf, int count)
{
  int i;

  for(i = 0; i < count; i++)
    buf[i] = random();
}

void
rwtest(char *filename)
{
  int max = 50000;
  int flen = 0; // Correct file length.
  char *fbytes; // Correct file contents.
  char *buf;
  int iters, fd;
  int start, count, cc, xx, i;

  fbytes = malloc(max);
  memset(fbytes, '\0', max);

  buf = malloc(max);

  unlink(filename);
  fd = open(filename, O_CREAT | O_TRUNC | O_RDWR, 0666);
  if(fd < 0){
    perror(filename);
    exit(1);
  }

  for(iters = 0; iters < 100; iters++){
    /*
     * Write
     */
    start = random() % max;
    count = random() % max;
    if(start + count > max)
      count = max - start;
    randomize(buf, count);
 
    if(lseek(fd, start, 0) != start){
      perror("lseek");
      unlink(filename);
      exit(1);
    }
    if(write(fd, buf, count) != count){
      perror("write");
      unlink(filename);
      exit(1);
    }

    memcpy(fbytes + start, buf, count);
    if(start + count > flen)
      flen = start + count;

    /*
     * Read
     */
    start = random() % max;
    count = random() % max;
    if(start + count > max)
      count = max - start;
    memset(buf, '\0', count);

    lseek(fd, start, 0);
    cc = read(fd, buf, count);
    xx = flen - start;
    if(xx > count)
      xx = count;
    if(xx < 0)
      xx = 0;
    if(cc != xx){
      fprintf(stderr, "rwtest cc %d xx %d start %d count %d flen %d\n",
              cc, xx, start, count, flen);
      unlink(filename);
      exit(1);
    }

    for(i = 0; i < cc; i++){
      if(buf[i] != fbytes[start+i]){
        fprintf(stderr, "mismatch iter %d at %d %02x vs %02x cc %d start %d count %d flen %d\n",
                iters, start+i, buf[i], fbytes[start+i], cc, start, count, flen);
        unlink(filename);
        exit(1);
      }
    }
  }

  close(fd);

  unlink(filename);
}
